@@ -3,7 +3,7 @@
 #   Dispatcher service for systemd-networkd
 # Copyright(c) 2016 by wave++ "Yuri D'Elia" <wavexx@thregr.org>
 # Distributed under GPLv3+ (see COPYING) WITHOUT ANY WARRANTY.
-# Copyright(c) 2018-2020 by craftyguy "Clayton Craft" <clayton@craftyguy.net>
+# Copyright(c) 2018-2022 by craftyguy "Clayton Craft" <clayton@craftyguy.net>
 # Distributed under GPLv3+ (see COPYING) WITHOUT ANY WARRANTY.
 
 
@@ -15,8 +15,9 @@ import errno
 import json
 import logging
 import os
+import pathlib
+import re
 import socket
-import stat
 import subprocess
 import sys
 
@@ -54,6 +55,11 @@ LOG_FORMAT = '%(levelname)s:%(message)s'
 
 SINGLETONS = {'Type', 'ESSID', 'OperationalState'}
 
+# taken from https://www.freedesktop.org/software/systemd/man/networkctl.html
+ADMIN_STATES = ['configured', 'configuring', 'failed', 'pending', 'unmanaged',
+                'linger', 'initialized']
+OPER_STATES = ['carrier', 'degraded', 'degraded-carrier', 'dormant',
+               'enslaved', 'missing', 'no-carrier', 'off', 'routable']
 
 AddressList = collections.namedtuple('AddressList', ['ipv4', 'ipv6'])
 NetworkctlListState = collections.namedtuple('NetworkctlListState',
@@ -61,6 +67,10 @@ NetworkctlListState = collections.namedtuple('NetworkctlListState',
                                               'operational', 'administrative'])
 
 
+class UnknownState(Exception):
+    pass
+
+
 def unquote(buf, char='\\'):
     """Remove escape characters from iwconfig ESSID output"""
     idx = 0
@@ -110,6 +120,10 @@ def get_networkctl_status(iface_name):
         if not v:
             continue
 
+        # normalize some values that changed in v244 & v246
+        if k == 'Address':
+            v = re.sub(r' \(DHCP4.*\)$', '', v)
+
         oldk = k
 
         if k in SINGLETONS:
@@ -144,11 +158,29 @@ def iw_get_ssid(iface_name):
 
 def iwconfig_get_ssid(iface_name):
     out = subprocess.check_output([IWCONFIG, '--', iface_name])
-    line = out.split(b'\n')[0].decode('utf-8', errors='replace')
+    line = out.split(b'\n', maxsplit=1)[0].decode('utf-8', errors='replace')
     essid = line[line.find('ESSID:')+7:-3]
     return unquote(essid)
 
 
+def check_perms(path, mode=0o755, uid=0, gid=0):
+    """ Check that the given file or dir @ path has the given mode set, and is
+    owned by the given uid/gid. Symlinks are *not* followed. Raises
+    FileNotFoundError if path doesn't exist."""
+
+    if not os.path.exists(path):
+        raise FileNotFoundError
+    st = os.stat(path, follow_symlinks=False)
+    st_mode = st.st_mode & 0x00FFF
+    if st.st_uid == uid and st.st_gid == gid and st_mode == mode:
+        return True
+
+    logger.error("invalid permissions on %s. expected mode=%s, uid=%d, "
+                 "gid=%d; got mode=%s, uid=%d, gid=%d", path, oct(mode), uid,
+                 gid, oct(st_mode), st.st_uid, st.st_gid)
+    return False
+
+
 def scripts_in_path(path, subdir):
     """Given directory names in PATH notation (separated by :), and a
     subdirectory name, return a sorted list of executables
@@ -167,20 +199,25 @@ def scripts_in_path(path, subdir):
     for filename in sorted(base_filenames):
         for one_path in path.split(":"):
             pathname = os.path.join(one_path, subdir, filename)
-            logger.debug("Checking if %s exists as %s", filename, pathname)
 
             if os.path.isfile(pathname):
-                entry = os.stat(pathname)
-                # Make sure script can be executed
-                if not stat.S_IXUSR & entry.st_mode:
-                    logger.error("Unable to execute script, check file mode: "
-                                 "%s", pathname)
-                # Make sure script is owned by root
-                elif entry.st_uid != 0 or entry.st_gid != 0:
-                    logger.error("Unable to execute script, check file perms: "
-                                 "%s", pathname)
-                else:
-                    script_list.append(pathname)
+                try:
+                    realpath = pathlib.Path(pathname).resolve()
+
+                    # Make sure that the file's parent dir has the correct
+                    # perms, without following any symlinks
+                    if not check_perms(os.path.dirname(pathname),
+                                       0o755, 0, 0):
+                        continue
+
+                    # Make sure file has correct perms, after following any
+                    # symlink(s)
+                    if not check_perms(realpath, 0o755, 0, 0):
+                        continue
+                except FileNotFoundError:
+                    continue
+
+                script_list.append(pathname)
                 break
 
     return script_list
@@ -229,14 +266,14 @@ class Dispatcher():
         self._interface_scan()
 
     def __repr__(self):
-        return '<Dispatcher(%r)>' % (self.__dict__,)
+        return f"<Dispatcher({self.__dict__})>"
 
     def _interface_scan(self):
         iface_list = get_networkctl_list()
         # Append new interfaces, keeping old ones around to avoid hotplug race
         # condition (issue #20)
         for i in iface_list:
-            if i not in self.iface_names_by_idx:
+            if i.idx not in self.iface_names_by_idx:
                 self.iface_names_by_idx[i.idx] = i.name
                 self.ifaces_by_name[i.name] = i
         logger.debug('Performed interface scan; state: %r', self)
@@ -262,10 +299,12 @@ class Dispatcher():
                                   administrative_state=iface.administrative,
                                   operational_state=iface.operational,
                                   force=True)
-            # pylint: disable=broad-except
-            except Exception:
-                logger.exception('Error handling initial for interface %r',
-                                 iface)
+            except UnknownState as e:
+                logger.exception("Unknown state for interface %s: %s",
+                                 iface, str(e))
+            except Exception:  # pylint: disable=broad-except
+                logger.exception("Error handling initial state for "
+                                 "interface %r", iface)
 
     def get_scripts_list(self, state):
         """Return scripts for the given state"""
@@ -301,6 +340,13 @@ class Dispatcher():
 
     def handle_state(self, iface_name, administrative_state=None,
                      operational_state=None, force=False):
+        if (administrative_state and
+                administrative_state.lower() not in ADMIN_STATES):
+            raise UnknownState(administrative_state)
+        if (operational_state and
+                operational_state.lower() not in OPER_STATES):
+            raise UnknownState(operational_state)
+
         self._handle_one_state(iface_name, administrative_state,
                                'administrative', force=force)
         self._handle_one_state(iface_name, operational_state, 'operational',
@@ -358,11 +404,13 @@ class Dispatcher():
         # http://thread.gmane.org/gmane.comp.sysutils.systemd.devel/36460
         idx = path[32:]
         idx = int(chr(int(idx[:2], 16)) + idx[2:])
+        is_new_interface = False
         if idx not in self.iface_names_by_idx:
             # Try to reload configuration if even an ignored message is seen
             logger.warning('Unknown index %r seen, reloading interface list',
                            idx)
             self._interface_scan()
+            is_new_interface = True
 
         try:
             iface_name = self.iface_names_by_idx[idx]
@@ -378,11 +426,16 @@ class Dispatcher():
 
         if ((operational_state is not None) or
                 (administrative_state is not None)):
-            self.handle_state(iface_name,
-                              administrative_state=str(administrative_state)
-                              if administrative_state else None,
-                              operational_state=str(operational_state)
-                              if operational_state else None,)
+            try:
+                self.handle_state(iface_name,
+                                  administrative_state=str(administrative_state)  # noqa
+                                  if administrative_state else None,
+                                  operational_state=str(operational_state)
+                                  if operational_state else None,
+                                  force=is_new_interface)
+            except UnknownState as e:
+                logger.exception("Unknown state for interface %s: %s",
+                                 iface_name, str(e))
 
         # Handle interfaces that have been removed
         if administrative_state == 'linger':
@@ -403,7 +456,7 @@ def sd_notify(**kwargs):
     try:
         # Turn state, a dictionary, into a properly formatted string where
         # each 'key=val' combo in the dictionary is separated by a \n
-        state_str = '\n'.join(['{0}={1}'.format(key, val) for (key, val)
+        state_str = '\n'.join([f"{key}={val}" for (key, val)
                                in kwargs.items()])
         env = os.environ.get('NOTIFY_SOCKET', None)
         if not env:
